import copy
from typing import Any, List, Callable, TypeVar, Generic, Optional, Set


__all__ = ("Proxy",)

T = TypeVar("T")


class _LazyCall:
    NULL_OBJECT = object()

    def __init__(self, fn: Callable[..., T]):
        self.fn = fn
        self.value = _LazyCall.NULL_OBJECT

    def __call__(self):
        if self.value is _LazyCall.NULL_OBJECT:
            self.value = self.fn()
        return self.value


def _error_func():
    """Function that is assigned by default, if no inner value
    or inner constructor supplied"""
    raise ValueError("There should be initialized constructor or obj.")


def _get_proxy_field(proxy: "Proxy[T]", name: str) -> Callable[..., T]:
    """Helper function to get proxy field"""
    return object.__getattribute__(proxy, name)


_CONSTRUCT_FIELD: str = "_value_provider"

_proxy_attrs: Set[str] = {
    "_set_provider",
    "_get_provider",
    _CONSTRUCT_FIELD,
    "__dict__",
}


class Proxy(Generic[T]):
    CONSTRUCT_FIELD = _CONSTRUCT_FIELD

    def __init__(
        self,
        value: T = None,
        value_provider: Callable[[], Optional[T]] = None,
        cached: bool = True,
    ):
        """
        Proxy class

        :param value: object that should be proxied
        :param value_provider: callable, that returns object that should be proxied
        :param cached: set False if you wants to call value_provider
            every time on access to proxy
        """
        if value and value_provider:
            raise ValueError("There should be one of obj or constructor, not both.")

        def _wrapper() -> Optional[T]:
            return value

        if value:
            value_provider = _wrapper
        elif value_provider:
            value_provider = _LazyCall(value_provider) if cached else value_provider
        else:
            value_provider = _error_func

        object.__setattr__(self, Proxy.CONSTRUCT_FIELD, value_provider)

    @staticmethod
    def set_value(proxy: "Proxy", value: Any, cached: bool = True) -> None:
        """Set inner value to proxy

        :param proxy: Proxy object
        :param value: value, that must be set
        :param cached: set False if you wants to call value_provider
            every time on access to proxy
        """
        proxy._set_provider(lambda: value, cached)

    @staticmethod
    def set_provider(proxy: "Proxy", provider: Callable, cached: bool = True) -> None:
        """Set callable, that on call return values to be proxied

        :param proxy: Proxy object
        :param provider: value provider that can produce proxied values
        :param cached: set False if you wants to call value_provider
            every time on access to proxy
        """
        proxy._set_provider(provider, cached)

    @staticmethod
    def get_value(proxy: "Proxy") -> Any:
        """Returns value from proxy"""
        return proxy._get_provider()()

    @staticmethod
    def is_initialized(proxy: "Proxy") -> bool:
        """Check, whether proxy is initialized with value or value constructor"""
        return proxy._get_provider() is not _error_func

    @staticmethod
    def set_proxies_values(proxies: List["Proxy"], values: List[Any]) -> None:
        """Set inner values to proxies.

        :param proxies: list of proxies, to set inner values
        :param values: list of inner values to set on proxies
        """
        if not len(proxies) == len(values):
            raise ValueError("Length of proxies and length or values must be equal")

        for proxy, value in zip(proxies, values):
            Proxy.set_value(proxy, value)

    @staticmethod
    def set_proxies_providers(
        proxies: List["Proxy"], providers: List[Callable], cached: bool = True
    ) -> None:
        """Set value providers to proxies.

        :param proxies: list of proxies, to set value providers
        :param providers: list of value providers to set on proxies
        :param cached: set False if you wants to call value_provider
            every time on access to proxy
        """
        if not len(proxies) == len(providers):
            raise ValueError("Length of proxies and length or values must be equal")

        if not all(callable(provider) for provider in providers):
            raise ValueError("Providers must be callable")

        for proxy, provider in zip(proxies, providers):
            Proxy.set_provider(proxy, provider, cached)

    __slots__ = [CONSTRUCT_FIELD, "__weakref__"]

    def _set_provider(self, value_provider: Callable, cached: bool = True) -> None:
        """Set value of proxy
        :param value_provider: Callable, that after return must return value,
            that will be stored in proxy
        :param cached: set False if you wants to call value_provider
            every time on access to proxy
        """
        provider = _LazyCall(value_provider) if cached else value_provider
        object.__setattr__(self, Proxy.CONSTRUCT_FIELD, provider)

    def _get_provider(self) -> Callable[..., T]:
        """Get value provider"""
        return _get_proxy_field(self, Proxy.CONSTRUCT_FIELD)

    @property
    def __dict__(self):
        try:
            return self._get_provider()().__dict__
        except ValueError:
            raise AttributeError("__dict__")

    def __dir__(self):
        try:
            return dir(self._get_provider()())
        except ValueError:
            return []

    def __setitem__(self, key, value):
        self._get_provider()()[key] = value

    def __delitem__(self, key):
        del self._get_provider()()[key]

    def __setattr__(self, key, value):
        return setattr(self._get_provider()(), key, value)

    def __delattr__(self, item):
        return delattr(self._get_provider()(), item)

    def __str__(self):
        return str(self._get_provider()())

    def __repr__(self):
        return repr(self._get_provider()())

    def __bytes__(self):
        return bytes(self._get_provider()())

    def __lt__(self, other):
        return self._get_provider()() < other

    def __le__(self, other):
        return self._get_provider()() <= other

    def __eq__(self, other):
        return self._get_provider()() == other

    def __ne__(self, other):
        return self._get_provider()() != other

    def __gt__(self, other):
        return self._get_provider()() > other

    def __ge__(self, other):
        return self._get_provider()() >= other

    def __hash__(self):
        return hash(self._get_provider()())

    def __call__(self, *args, **kwargs):
        return self._get_provider()()(*args, **kwargs)

    def __len__(self):
        return len(self._get_provider()())

    def __getitem__(self, item):
        return self._get_provider()()[item]

    def __iter__(self):
        return iter(self._get_provider()())

    def __next__(self):
        return next(self._get_provider()())

    def __contains__(self, item):
        return item in self._get_provider()()

    def __add__(self, other):
        return self._get_provider()() + other

    def __sub__(self, other):
        return self._get_provider()() - other

    def __mul__(self, other):
        return self._get_provider()() * other

    def __floordiv__(self, other):
        return self._get_provider()() // other

    def __mod__(self, other):
        return self._get_provider()() % other

    def __divmod__(self, other):
        return divmod(self._get_provider()(), other)

    def __pow__(self, power):
        return self._get_provider()() ** power

    def __lshift__(self, other):
        return self._get_provider()() << other

    def __rshift__(self, other):
        return self._get_provider()() >> other

    def __and__(self, other):
        return self._get_provider()() & other

    def __xor__(self, other):
        return self._get_provider()() ^ other

    def __or__(self, other):
        return self._get_provider()() | other

    def __truediv__(self, other):
        return self._get_provider()() / other

    def __neg__(self):
        return -(self._get_provider()())

    def __pos__(self):
        return +(self._get_provider()())

    def __abs__(self):
        return abs(self._get_provider()())

    def __invert__(self):
        return ~(self._get_provider()())

    def __complex__(self):
        return complex(self._get_provider()())

    def __int__(self):
        return int(self._get_provider()())

    def __float__(self):
        return float(self._get_provider()())

    def __index__(self):
        return self._get_provider()().__index__()

    def __enter__(self):
        return self._get_provider()().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._get_provider()().__exit__(exc_type, exc_val, exc_tb)

    def __radd__(self, other):
        return other + self._get_provider()()

    def __rsub__(self, other):
        return other - self._get_provider()()

    def __rmul__(self, other):
        return other * self._get_provider()()

    def __rtruediv__(self, other):
        return other / self._get_provider()()

    def __rfloordiv__(self, other):
        return other // self._get_provider()()

    def __rmod__(self, other):
        return other % self._get_provider()()

    def __rdivmod__(self, other):
        return divmod(other, self._get_provider()())

    def __copy__(self):
        return copy.copy(self._get_provider()())

    def __round__(self, n=None):
        return round(self._get_provider()(), n)

    def __trunc__(self):
        return self._get_provider()().__trunc__()

    def __reversed__(self):
        return reversed(self._get_provider()())

    def __rpow__(self, other):
        return other ** self._get_provider()()

    def __rlshift__(self, other):
        return other << self._get_provider()()

    def __rrshift__(self, other):
        return other >> self._get_provider()()

    def __rand__(self, other):
        return other & self._get_provider()()

    def __rxor__(self, other):
        return other ^ self._get_provider()()

    def __ror__(self, other):
        return other | self._get_provider()()

    def __getattribute__(self, name):
        if name in _proxy_attrs:
            return _get_proxy_field(self, name)

        return getattr(_get_proxy_field(self, Proxy.CONSTRUCT_FIELD)(), name)

    def __bool__(self) -> bool:
        try:
            try:
                return getattr(
                    _get_proxy_field(self, Proxy.CONSTRUCT_FIELD)(), "__bool__"
                )()
            except AttributeError:
                pass
            try:
                return bool(
                    getattr(_get_proxy_field(self, Proxy.CONSTRUCT_FIELD)(), "__len__")()
                )
            except AttributeError:
                return True
        except ValueError:
            return False



# -*- coding: utf-8 -*-
from setuptools import setup

modules = \
['proxy']
setup_kwargs = {
    'name': 'proxied',
    'version': '1.0.1',
    'description': 'A simple to use proxy for python objects',
    'long_description': 'proxied\n-------\nproxied package can be use to defer initialization of object.\n\nSuppose we have:\n\n**__init__.py**\n\n.. code-block:: python\n\n    from config import Config, load_config\n\n    config: Config = None\n\n    def init_app():\n        config = load_config()\n\nIf we try to import config in another file, i.e.\n\n**cool_staff.py**\n\n.. code-block:: python\n\n    from . import config\n\n    ...\n\n    init_staff(config.db_url)\n\nWe will end up importing None. So, we will be forced to\nmake import not at the top of the file.\n\n**Here comes the proxied package**\n\n**__init__.py**\n\n.. code-block:: python\n\n    from typing import Union\n    from proxy import Proxy\n    from config import Config, load_config\n\n    config: Union[Config, Proxy] = Proxy()\n\n    def init_app():\n        config.set_inner(load_config())\n\nNow we can easily import config at the top of the file work with\nproxy object as it was the object of type Config.\n\nInstallation\n------------\n\n.. code-block::\n\n    pip install proxied\n\nProxy\n-----------------------\nProxy class can be initialized with inner or inner_constructor.\nIf inner_constructor is supplied, then it will be called once,\nand the result will be cached.\n\nIf Proxy class is initialized without inner and inner_constructor,\nthe inner should be set later with a help of ``proxy.set_inner`` method.\n\n.. code-block:: python\n\n    from proxy import Proxy\n    proxy = Proxy()\n    proxy.set_inner({})\n    proxy["test_key"] = 10\n\n\nIt\'s possible to set values for multiple proxies.\n\n.. code-block:: python\n\n    from proxy import Proxy\n    proxies = [Proxy(), Proxy(), Proxy(), Proxy()]\n    values = [10, 11, list(), dict()]\n    Proxy.set_proxies(proxies, values)\n\nThere is a check, if proxy is initialized with proxied value\n\n.. code-block:: python\n\n    from proxy import Proxy\n    proxy = Proxy()\n\n    if not proxy.initialized:\n        data = get_needed_data(...)\n        proxy.set_inner(data)\n\nExample\n-------\n.. code-block:: python\n\n    from proxy import Proxy\n    class NotAvailableDuringImport:\n        @property\n        def data(self):\n            return "Not Available during import"\n\n\n    proxy: Union[NotAvailableDuringImport, Proxy] = Proxy()\n    proxy.set_inner(NotAvailableDuringImport)\n    assert proxy.data == "Not Available during import"\n\nLicense\n-------\n\nCopyright Oleksii Petrenko, 2020.\n\nDistributed under the terms of the `MIT`_ license,\njson_modify is free and open source software.\n\n.. _`MIT`: https://github.com/Enacero/proxied/blob/master/LICENSE',
    'author': 'Oleksii Petrenko',
    'author_email': 'oleksiiypetrenko@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Enacero/proxied',
    'py_modules': modules,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
